#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <pc.h>
#include <kb.h>
#include <bios.h>
#include <sys/movedata.h>
#include "blin.h"
#include "yuck.h"

#define VecList(V) V[0], V[1], V[2]

#define VecLen2(V) sqrt(VecDot2(V, V))
#define VecList2(V) V[0], V[1]
#define VecDot2(A, B) ((A[0]*B[0]) + (A[1]*B[1]))

#define VecScalar2(V, assign_op, k) \
(	V[0] assign_op k, \
	V[1] assign_op k \
)
#define VecVec2(A, assign_op, B) \
(	A[0] assign_op B[0], \
	A[1] assign_op B[1] \
)
#define VecVecVec2(A, assign_op, B, op, C) \
(	A[0] assign_op B[0] op C[0], \
	A[1] assign_op B[1] op C[1] \
)
#define MatVec2(A, M, B) \
(	A[0] = M[0][0]*B[0] + M[0][1]*B[1], \
	A[1] = M[1][0]*B[0] + M[1][1]*B[1] \
)

#define epsilon 1e-10
/*
#define appendwall(a, b) \
	a->next = (wlelem *) xmalloc(sizeof(wlelem)); \
	a = a->next; \
	a->next = NULL; \
	a->w = b;
	*/

#define texel(n, tx, ty) tex[n].data[ty * tex[n].width + tx]

#define angleinc M_PI / 40

#define frontclip 5.0

enum {
	string_max = 80,
	debug = 0
};

typedef double vector[3];
typedef double matrix[3][3];
typedef double vector2[2];
typedef double matrix2[2][2];

struct wallattr {
	//double red, green, blue;
	int texture;
	short visible;
};

typedef struct wallattr wallattr;

struct wall {
	vector2 start;
	vector2 end;
	vector2 n;
	double length;
	wallattr plus, minus;
};

typedef struct wall wall;

struct wlelem {
	struct wlelem *next;
	wall w;
};

typedef struct wlelem wlelem;

struct texture_map {
	int width;
	int height;
	int logheight;
	int wmask;
	int hmask;
	byte *data;
	char filename[string_max];
};

typedef struct texture_map texture_map;

struct plane {
	vector2 p;
	vector2 n;
};

typedef struct plane plane;

struct bsptreenode {
	struct bsptreenode *front;
	struct bsptreenode *back;
	plane cut;
	wlelem *wl;
};

typedef struct bsptreenode bsptreenode;

struct sceneinfo {
	vector eye;
	vector at;
	vector up;
	double focal;
	int swidth;
	int sheight;
	double scale;
	vector2 leftn, rightn;
};

typedef struct sceneinfo sceneinfo;

// prototypes
void savefile(char *filename);
void showtexture(int x, int y, texture_map *t);
void painter(bsptreenode *node);
void drawwall(wall *w, char);
void drawview(void);
void drawmap(void);
void handlekey(int key);

void inorderdrawmap(bsptreenode *r);
void drawscene(void);
void plot(int x, int y, byte c);
void bresenham(int x1, int y1, int x2, int y2, byte c);
void printwall(wall *);
void preorderprint(bsptreenode *);
void printwalllist(wlelem *);
void makebsptree(bsptreenode *, wlelem *);
void readin(char *, void *, char *, ...);
void parsecomment(void);
void loadtex(int n, char *filename);
void loadfile(char *);

void *xmalloc(size_t);
void *normalize2(double [2]);
void readinwallattr(wallattr *);

int wallcount;
wlelem *wlhead, *origwlhead;
bsptreenode *bsproot;

char stopflag;
char mapmode;

FILE *infile;

sceneinfo sc;
//double angleinc = M_PI / 32;
double sinangleinc, cosangleinc;
double stepsize = 2;

//matrix2 transform;

char img[320 * 200];
texture_map *tex;
int texcount;

int ceilheight = 128;
int floorheight = 0;
int eyeheight = 64;

int main(int argc, char *argv[]) {
	int i;
	int x1, y1, x2, y2;
	wlelem *we;
	/*
	window w;
	drawing d;
	*/

	if (argc < 2) {
		fprintf(stderr, "No filename given\n");
		exit(2);
	}

	loadfile(argv[1]);
	savefile("temp");

	cosangleinc = cos(angleinc);
	sinangleinc = sin(angleinc);

	bsproot = (bsptreenode *) xmalloc(sizeof(bsptreenode));
	makebsptree(bsproot, wlhead);

	/*
	printf("BSP tree:\n");
	preorderprint(bsproot);
	
	printf("End of BSP tree\n");
	printf("%d\n", wallcount);

	we = wlhead->next;

	while(we) {
		printf("%lf %lf %lf %lf\n", we->w.start[0], we->w.start[1],
				we->w.end[0], we->w.end[1]);
		printf("%lf %lf\n", we->w.n[0], we->w.n[1]);
		we = we->next;
	}
	*/

	initgraphics();

	sc.swidth = 320;
	sc.sheight = 200;

	sc.focal = 200;
	sc.scale = 16;
	sc.rightn[0] = -(sc.leftn[0] = sc.focal);
	sc.rightn[1] = sc.leftn[1] = sc.swidth / 2.0;

	logprintf("focal: %lf\tscale: %lf", sc.focal, sc.scale);

	mapmode = 1;
	stopflag = 0;
	kb_install(0);

	// Main Loop
	while(!stopflag) {
		drawscene();
		kb_update();
		for (i=0; i<128; i++) {
			if (kb_key(i)) handlekey(i);
		}
	}

	settextmode();
}

void handlekey(int key) {
	double x, y;

	switch(key) {
		case KB_SCAN_TAB:
			mapmode = !mapmode;
			delay(50);
			break;
		case KB_SCAN_S:
			if (kb_shift() & KB_SHIFT_ANY_ALT) {
				sc.scale /= 2;
			} else {
				sc.scale *= 2;
			}
			break;
		case KB_SCAN_E:
			if (kb_shift() & KB_SHIFT_ANY_ALT) {
				eyeheight-=stepsize;
			} else {
				eyeheight+=stepsize;
			}
			break;
		case KB_SCAN_1:
			sc.focal = 1;
			sc.rightn[0] = -(sc.leftn[0] = sc.focal);
			break;
		case KB_SCAN_F:
			if (kb_shift() & KB_SHIFT_ANY_ALT) {
				sc.focal /= 2;
				sc.rightn[0] = -(sc.leftn[0] = sc.focal);
			} else {
				sc.focal *= 2;
				sc.rightn[0] = -(sc.leftn[0] = sc.focal);
			}
			break;
		case KB_SCAN_C:
			if (kb_shift() & KB_SHIFT_ANY_ALT) {
				ceilheight /= 2;
				eyeheight /= 2;
			} else {
				ceilheight *= 2;
				eyeheight *= 2;
			}
			break;
		case KB_SCAN_Q:
			stopflag = 1;
			break;
		case KB_SCAN_4_PAD:
		case KB_SCAN_LEFT:
			if (kb_shift() & KB_SHIFT_ANY_ALT) {
				x = -sc.at[1];
				y = sc.at[0];
				sc.eye[0] += stepsize * x;
				sc.eye[1] += stepsize * y;
			} else {
				x = sc.at[0];
				y = sc.at[1];
				sc.at[0] = cosangleinc * x - sinangleinc * y;
				sc.at[1] = sinangleinc * x + cosangleinc * y;
			}
			break;
		case KB_SCAN_6_PAD:
		case KB_SCAN_RIGHT:
			if (kb_shift() & KB_SHIFT_ANY_ALT) {
				x = sc.at[1];
				y = -sc.at[0];
				sc.eye[0] += stepsize * x;
				sc.eye[1] += stepsize * y;
			} else {
				x = sc.at[0];
				y = sc.at[1];
				sc.at[0] = cosangleinc * x + sinangleinc * y;
				sc.at[1] = -sinangleinc * x + cosangleinc * y;
			}
			break;
		case KB_SCAN_8_PAD:
		case KB_SCAN_UP:
			VecVec2(sc.eye, +=, stepsize * sc.at);
			break;
		case KB_SCAN_2_PAD:
		case KB_SCAN_DOWN:
			VecVec2(sc.eye, -=, stepsize * sc.at);
			break;
		default:
			break;
	}
}

void drawscene(void) {
	int i, j;

	for (i=0; i<(sc.swidth * sc.sheight); i++) img[i] = 0;

	drawview();
	if (mapmode) {
		drawmap();
	}

	drawimage(img);
}

void drawview(void) {
	painter(bsproot);
}

void painter(bsptreenode *node) {
	wlelem *we;
	vector2 p;
	double d;

	VecVecVec2(p, =, sc.eye, -, node->cut.p);
	d = VecDot2(node->cut.n, p);

	if (d > epsilon) {
		if (node->back) {
			painter(node->back);
		}

		we = node->wl->next;
		while(we) {
			if (we->w.plus.visible) {
				drawwall(&we->w, 1);
			}
			we = we->next;
		}

		if (node->front) {
			painter(node->front);
		}
	} else if (d < epsilon) {
		if (node->front) {
			painter(node->front);
		}
		we = node->wl->next;
		while(we) {
			if (we->w.minus.visible) {
				drawwall(&we->w, -1);
			}
			we = we->next;
		}
		if (node->back) {
			painter(node->back);
		}
	}
}

void drawwall(wall *w, char side) {
	int x1, x2;
	int y1t, y1b, y2t, y2b;
	int dx, dy;
	int dyt, dyb;
	int errort, errorb;
	int yinct, yincb;
	int i, j;
	wallattr *wa;
	matrix2 m;
	vector2 v, v1, v2;
	vector2 nv1, nv2;
	vector2 dir;
	double mu1, mu2;
	double d;
	double recip1, recip2, drecip;
	int ymin, ymax;
	double xmult;
	int tx, ty;
	int dty;
	int hmask, wmask;
	byte *texdata;
	int texlogheight;
	double txstart;
	int temp;

	m[0][0] = sc.scale * sc.at[1];
	m[0][1] = sc.scale * -sc.at[0];
	m[1][0] = sc.scale * sc.at[0];
	m[1][1] = sc.scale * sc.at[1];

	v[0] = w->start[0] - sc.eye[0];
	v[1] = w->start[1] - sc.eye[1];
	MatVec2(v1, m, v);

	v[0] = w->end[0] - sc.eye[0];
	v[1] = w->end[1] - sc.eye[1];
	MatVec2(v2, m, v);

	logprintf("\nTesting points...\n");

	VecVecVec2(dir, =, v2, -, v1);

	if(0) {
		//WRONG! Forgot to transform w->n
	/* v is the normal rotated 90 degrees */
	v[0] = -w->n[1] * side;
	v[1] = w->n[0] * side;
	/* if v and dir face different ways then swap v1 and v2 */
	if (VecDot2(v, dir) < -epsilon) {
		VecVec2(v, =, v1);
		VecVec2(v1, =, v2);
		VecVec2(v2, =, v);
	}
	}

	logprintf("rotated normal: %lf, %lf\n", v[0], v[1]);

	if (side == -1) {
		VecVec2(v, =, v1);
		VecVec2(v1, =, v2);
		VecVec2(v2, =, v);
		VecVec2(dir, =, -dir);
	}
	logprintf("at: %lf, %lf, side: %d\n", sc.at[0], sc.at[1], side);
	logprintf("%lf, %lf -> %lf, %lf\n", v1[0], v1[1], v2[0], v2[1]);
	
	d = VecDot2(sc.leftn, dir);
	VecVec2(nv1, =, v1);
	if (fabs(d) > epsilon) {
		mu1 = -VecDot2(sc.leftn, v1) / d;
		logprintf("mu: %lf\n", mu1);
		if (mu1 > 0 && mu1 <= 1 && (d = v1[1] + mu1*dir[1]) > 0) {
			nv1[0] = v1[0] + mu1*dir[0];
			nv1[1] = d;
			logprintf("Left Intersection %lf, %lf\n", VecList2(nv1));
			logprintf("should be 0: %lf\n", VecDot2(sc.leftn, nv1));
		}
	}

	d = VecDot2(sc.rightn, dir);
	VecVec2(nv2, =, v2);
	if (fabs(d) > epsilon) {
		mu2 = -VecDot2(sc.rightn, v1) / d;
		logprintf("mu: %lf\n", mu2);
		if (mu2 > 0 && mu2 <= 1 && (d = v1[1] + mu2*dir[1]) > 0) {
			nv2[0] = v1[0] + mu2*dir[0];
			nv2[1] = d;
			logprintf("Right Intersection %lf, %lf\n", VecList2(nv2));
			logprintf("should be 0: %lf\n", VecDot2(sc.rightn, nv2));
		}
	}

	logprintf("clipped: %lf, %lf -> %lf, %lf\n", nv1[0], nv1[1], nv2[0], nv2[1]);
	/* if either point is outside, don't draw it */
	if (nv1[1] < frontclip || nv1[1]+(sc.leftn[0]/sc.leftn[1])*nv1[0] < -epsilon
			|| nv1[1]+(sc.rightn[0]/sc.rightn[1])*nv1[0] < -epsilon) {
		logprintf("nv1 outside\n");
		return;
	}
	if (nv2[1] < frontclip || nv2[1]+(sc.leftn[0]/sc.leftn[1])*nv2[0] < -epsilon
			|| nv2[1]+(sc.rightn[0]/sc.rightn[1])*nv2[0] < -epsilon) {
		logprintf("nv2 outside\n");
		return;
	}

	if (side == 1) {
		wa = &w->plus;
	} else {
		wa = &w->minus;
	}

	wmask = tex[wa->texture].wmask;
	hmask = tex[wa->texture].hmask;
	texdata = tex[wa->texture].data;
	texlogheight = tex[wa->texture].logheight;

	logprintf("at: %lf, %lf\n", sc.at[0], sc.at[1]);

	x1 = floor(sc.focal * nv1[0] / nv1[1]) + sc.swidth / 2;
	if (x1 < 0) x1 = 0;

	x2 = (int) (sc.focal * nv2[0] / nv2[1]) + sc.swidth / 2;
	if (x2 >= sc.swidth) x2 = sc.swidth - 1;

	if (x1 > x2) printf("BUG!\n");
	if (x1 == x2) return;
	logprintf("%d %d (x)\n", x1, x2);
	y1t = sc.sheight / 2 - (int) (sc.focal * (ceilheight - eyeheight) / nv1[1]);
	y1b = sc.sheight / 2 - (int) (sc.focal * (floorheight - eyeheight) / nv1[1]);
	y2t = sc.sheight / 2 - (int) (sc.focal * (ceilheight - eyeheight) / nv2[1]);
	y2b = sc.sheight / 2 - (int) (sc.focal * (floorheight - eyeheight) / nv2[1]);

	/*
	bresenham(x1, y1t, x2, y2t, wa->texture);
	bresenham(x1, y1b, x2, y2b, wa->texture);
	bresenham(x1, y1t, x1, y1b, wa->texture);
	bresenham(x2, y2t, x2, y2b, wa->texture);
	*/

	dx = x2 - x1;
	d = fabs(v2[1] - v1[1]);

	if (d < epsilon) {
		//special, easy case
		xmult = (nv2[0] - nv1[0]) / dx;
		dy = y1b - y1t;
		if (!dy) return;
		ymin = (y1t < 0) ? 0 : y1t;
		ymax = (y1b >= sc.sheight) ? sc.sheight - 1 : y1b;
		temp = (ymax - ymin) * (ceilheight - floorheight) / dy;
		errorb = -(ymax - ymin) / 2;
		for (i=x1; i<=x2; i++) {
			tx = (int) (xmult * (i - x1) + nv1[0] - v1[0]) & wmask;
			ty = ((ceilheight - floorheight) * (ymin - y1t) / dy) & hmask;
			errort = errorb;
			for (j=ymin; j<=ymax; j++) {
				//ty = ((ceilheight - floorheight) * (j - y1t) / dy) & hmask;
				plot(i, j, texdata[(ty << texlogheight) + tx]);
				errort += temp;
				while (errort > 0) {
					ty++;
					ty&=hmask;
					errort -= (ymax - ymin);
				}
			}

		}
	} else {

		/* optimize by unrolling these cases? */
		// (= 4x more code)
		dyt = y2t - y1t;
		if (dyt < 0) {
			yinct = -1;
			dyt = -dyt;
		} else {
			yinct = 1;
		}
		dyb = y2b - y1b;
		if (dyb < 0) {
			yincb = -1;
			dyb = -dyb;
		} else {
			yincb = 1;
		}

		recip1 = 1/nv1[1];
		recip2 = 1/nv2[1];

		drecip = (recip2 - recip1) / dx;

		errorb = errort = -dx / 2;

		xmult = sc.scale * w->length / d;
		dy = y1b - y1t;
		ymin = (y1t < 0) ? 0 : y1t;
		ymax = (y1b >= sc.sheight) ? sc.sheight - 1 : y1b;
		txstart = (v1[1] * xmult);
		for (i=x1; i<=x2; i++) {

			if (dy) {
				temp =  ((ceilheight - floorheight) << 13) / dy;
				tx = abs(xmult / recip1 - txstart) & wmask;
				for (dty=ymin-y1t, j=i+ymin*sc.swidth; dty<ymax-y1t;
						dty++, j+=sc.swidth) {
					//ty = ((ceilheight - floorheight) * dty / dy) & hmask;
					ty = ((temp * dty) >> 13) & hmask;
					img[j] = texdata[tx + (ty << texlogheight)];
				}
			}

			recip1 += drecip;
			errort += dyt;
			while (errort > 0 && y1t != y2t) {
				y1t += yinct;
				dy -= yinct;
				errort -= dx;
				ymin = (y1t < 0) ? 0 : y1t;
			}
				
			errorb += dyb;
			while (errorb > 0 && y1b != y2b) {
				y1b += yincb;
				dy += yincb;
				errorb -= dx;
				ymax = (y1b > sc.sheight) ? sc.sheight - 1 : y1b;
			}
		}
	}
}

void drawmap(void) {
	byte c;
	int x1, y1, x2, y2;
	
	c = 14;
	x1 = sc.eye[0];
	y1 = sc.sheight - sc.eye[1];
	x2 = x1 + sc.at[0] * 3;
	y2 = y1 - sc.at[1] * 3;
	bresenham(x1, y1, x2, y2, c);
	c = 15;
	plot(x1, y1, c);
	inorderdrawmap(bsproot);
}

void inorderdrawmap(bsptreenode *r) {
	wlelem *we;
	byte c;
	int x1, x2, y1, y2;
	
	if (r->front) {
		inorderdrawmap(r->front);
	}
	if (r->back) {
		inorderdrawmap(r->back);
	}

	c = 15;

	we = r->wl->next;
	while (we) {
		x1 = (int) we->w.start[0];
		y1 = sc.sheight - (int) we->w.start[1];
		x2 = (int) we->w.end[0];
		y2 = sc.sheight - (int) we->w.end[1];
		bresenham(x1, y1, x2, y2, c);
		x1 = (x1 + x2) / 2;
		y1 = (y1 + y2) / 2;
		if (we->w.plus.visible)
			bresenham(x1, y1, x1 + (int) we->w.n[0] * 10,
					y1 - (int) we->w.n[1] * 10, we->w.plus.texture);
		if (we->w.minus.visible)
			bresenham(x1, y1, x1 - (int) we->w.n[0] * 10,
					y1 + (int) we->w.n[1] * 10, we->w.minus.texture);
		we = we->next;
	}
}

void savefile(char *filename) {
	int i, j, k;
	wlelem *we;
	wall *wp;
	vector2 d;
	FILE *fp;

	if (!(fp = fopen(filename, "w"))) {
			perror("fopen");
			exit(1);
	}

	fprintf(fp, "// Scene file\n");
	fprintf(fp, "\n// number of textures\n%d", texcount);
	fprintf(fp, "\n// Texture filenames:\n");
	for (i=0; i<texcount; i++) {
		fprintf(fp, "%s\n", tex[i].filename);
	}

	fprintf(fp, "\n// number of walls\n%d\n\n", wallcount);

	/*
	wlhead = (wlelem *) xmalloc(sizeof(wlelem));
	wlhead->next = NULL;

	we = wlhead;
	for (i=0; i<wallcount; i++) {
		we->next = (wlelem *) xmalloc(sizeof(wlelem));
		we = we->next;
		we->next = NULL;
		wp = &we->w;
		readin("%lf", &wp->start[0], "wall #%d start x", i);
		readin("%lf", &wp->start[1], "wall #%d start y", i);
		readin("%lf", &wp->end[0], "wall #%d end x", i);
		readin("%lf", &wp->end[1], "wall #%d end y", i);
		readin("%lf", &wp->n[0], "wall #%d normal x", i);
		readin("%lf", &wp->n[1], "wall #%d normal y", i);
		VecVecVec2(d, =, wp->end, -, wp->start);
		wp->length = sqrt(VecDot2(d, d));
		readinwallattr(&wp->plus);
		readinwallattr(&wp->minus);
	}

	fprintf(fp, "%d %s\n", sc.eye[0], "player start x");
	fprintf(fp, "%d %s\n", sc.eye[1], "player start y");
	*/

	sc.eye[2] = eyeheight;
	sc.at[0] = 1;
	sc.at[1] = 0;
}

void readinwallattr(wallattr *wattr) {
	int i;

	readin("%hd", &wattr->visible, "visibility flag");

	if (wattr->visible) {
		readin("%d", &wattr->texture, "wall #%d texture", i);
		/*
		readin("%lf", &wattr->c[0], "wall #%d side red", i);
		readin("%lf", &wattr->c[1], "wall #%d side green", i);
		readin("%lf", &wattr->c[2], "wall #%d side blue", i);
		*/
	}
}

void loadfile(char *filename) {
	int i, j, k;
	wlelem *we;
	wall *wp;
	vector2 d;
	FILE *fp;

	if (!(infile = fopen(filename, "r"))) {
			perror("fopen");
			exit(1);
	}

	readin("%d", &texcount, "number of textures");
	for (i=0; i<texcount; i++) {
		readin("%s", tex[i].filename, "texture filename");
		fp = infile;
		loadtex(i, tex[i].filename);
		infile = fp;
	}

	readin("%d", &wallcount, "number of walls");

	wlhead = (wlelem *) xmalloc(sizeof(wlelem));
	wlhead->next = NULL;

	we = wlhead;
	for (i=0; i<wallcount; i++) {
		we->next = (wlelem *) xmalloc(sizeof(wlelem));
		we = we->next;
		we->next = NULL;
		wp = &we->w;
		readin("%lf", &wp->start[0], "wall #%d start x", i);
		readin("%lf", &wp->start[1], "wall #%d start y", i);
		readin("%lf", &wp->end[0], "wall #%d end x", i);
		readin("%lf", &wp->end[1], "wall #%d end y", i);
		readin("%lf", &wp->n[0], "wall #%d normal x", i);
		readin("%lf", &wp->n[1], "wall #%d normal y", i);
		VecVecVec2(d, =, wp->end, -, wp->start);
		wp->length = sqrt(VecDot2(d, d));
		readinwallattr(&wp->plus);
		readinwallattr(&wp->minus);
	}

	readin("%lf", &sc.eye[0], "player start x");
	readin("%lf", &sc.eye[1], "player start y");
	sc.eye[2] = eyeheight;
	sc.at[0] = 1;
	sc.at[1] = 0;
}

#if 0
void savewallattr(wallattr *wattr) {
	int i;

	readin("%hd", &wattr->visible, "visibility flag");

	if (wattr->visible) {
		readin("%d", &wattr->texture, "wall #%d texture", i);
		/*
		readin("%lf", &wattr->c[0], "wall #%d side red", i);
		readin("%lf", &wattr->c[1], "wall #%d side green", i);
		readin("%lf", &wattr->c[2], "wall #%d side blue", i);
		*/
	}
}
#endif

void readin(char *s, void *f, char *format, ...) {
	va_list arg;
	char done;
	int check;
	char ch;

	done = 0;
	do {
		if (feof(infile)) {
			va_start(arg, format);
			fprintf(stderr, "Unexpected end of file.\nError reading ");
			vfprintf(stderr, format, arg);
			fprintf(stderr, "\n");
			va_end(arg);
			exit(1);
		}

		fscanf(infile, " %c", &ch);
		if (ch == '/') {
			parsecomment();
		} else {
			ungetc(ch, infile);
			check = fscanf(infile, s, f);
			if (!check) {
				va_start(arg, format);
				fprintf(stderr, "Error reading ");
				vfprintf(stderr, format, arg);
				fprintf(stderr, "\n");
				va_end(arg);
				exit(1);
			}
			//printf("Read %s\n", thing);
			done++;
		}
	} while (!done);
}

void parsecomment(void) {
	char ch;
	char flag;

	if ((ch = getc(infile) == '/')) {
		while ((ch = getc(infile) != '\n'));
		return;
	} else if ((ch = getc(infile)) != '*') {
		fprintf(stderr, "Warning: bad comment\n");
	}
	flag = 0;
	do {
		do {
			ch = getc(infile);
		} while (ch != '*');
		if ((ch = getc(infile)) == '/') {
			flag++;
		}
	} while (!flag);
}

void *normalize2(double v[2]) {
	double n;

	n = VecLen2(v); //sqrt(v[0] * v[0] + v[1] * v[1]);
	VecScalar2(v, /=, n);
	/*
	v[0] /= n;
	v[1] /= n;
	*/
}

void makebsptree(bsptreenode *node, wlelem *wl) {
	wlelem *wlfronthead, *wlbackhead;
	wlelem *wefront, *weback, *weon, *we;
	wlelem *newwe;
	double d, d2, mu;
	vector2 p, dir, v;
	char type;

	wefront = wlfronthead = (wlelem *) xmalloc(sizeof(wlelem));
	weback = wlbackhead = (wlelem *) xmalloc(sizeof(wlelem));
	wefront->next = weback->next = NULL;
	node->wl = (wlelem *) xmalloc(sizeof(wlelem));
	node->wl->next = NULL;

	weon = node->wl;

	we = wl->next;

	/*
	assert(we);
	if (!we) {
		return;
	}

	printf("Cut: ");
	printwall(&we->w);
	*/

	VecVec2(node->cut.p, =, we->w.start);
	VecVec2(node->cut.n, =, we->w.n);
	/*
	weon->next = (wlelem *) xmalloc(sizeof(wlelem));
	weon = weon->next;
	weon->next = NULL;
	weon->w = we->w;
	*/
	weon->next = we;
	weon = we;
	we = we->next;
	weon->next = NULL;

	while(we) {
		VecVecVec2(p, =, we->w.start, -, node->cut.p);
		d = VecDot2(node->cut.n, p);
		VecVecVec2(p, =, we->w.end, -, node->cut.p);
		d2 = VecDot2(node->cut.n, p);
		// 1 = in front, -1 = behind, 0 = on, 2 = cross
		if (d > epsilon) {
			if (d2 >= -epsilon) type = 1;
			else type = 2;
		} else if (fabs(d) < epsilon) {
			if (d2 > epsilon) type = 1;
			else if (d2 < -epsilon) type = -1;
			else type = 0;
		} else {
			if (d2 <= epsilon) type = -1;
			else type = 2;
		}
		switch(type) {
			case 0:
				weon->next = we;
				weon = we;
				we = we->next;
				weon->next = NULL;
				/*
				weon->next = (wlelem *) xmalloc(sizeof(wlelem));
				weon = weon->next;
				weon->next = NULL;
				weon->w = we->w;
				*/
				break;

			case 1:
				wefront->next = we;
				wefront = we;
				we = we->next;
				wefront->next = NULL;
				break;
				
			case -1:
				weback->next = we;
				weback = we;
				we = we->next;
				weback->next = NULL;
				break;

			case 2:
				// cut the wall
				// intersection of 2 lines
				VecVecVec2(dir, =, we->w.end, -, we->w.start);
				VecVecVec2(v, =, node->cut.p, -, we->w.start);
				mu = VecDot2(node->cut.n, v) / VecDot2(node->cut.n, dir);
				VecVecVec2(p, =, we->w.start, +, mu*dir);
				VecVecVec2(v, =, node->cut.p, -, p);
				printf("Intersection %lf, %lf\n", VecList2(p));
				printf("should be 0: %lf\n", VecDot2(node->cut.n, v));
				
				newwe = (wlelem *) xmalloc(sizeof(wlelem));
				newwe->w = we->w;
				VecVec2(newwe->w.start, =, p);
				VecVec2(we->w.end, =, p);
				if (d > epsilon) {
					wefront->next = we;
					wefront = we;
					we = we->next;
					wefront->next = NULL;
					weback->next = newwe;
					weback = newwe;
					weback->next = NULL;
				} else {
					weback->next = we;
					weback = we;
					we = we->next;
					weback->next = NULL;
					wefront->next = newwe;
					wefront = newwe;
					wefront->next = NULL;
				}
				break;

			default:
				we = we->next;
				break;
		}
	
	}
	if (wlfronthead->next) {
		node->front = (bsptreenode *) xmalloc(sizeof(bsptreenode));
		makebsptree(node->front, wlfronthead);
	} else {
		node->front = NULL;
	}
	if (wlbackhead->next) {
		node->back = (bsptreenode *) xmalloc(sizeof(bsptreenode));
		makebsptree(node->back, wlbackhead);
	} else {
		node->back = NULL;
	}
}

void preorderprint(bsptreenode *n) {
	static int lev = 0;
	int i;

	assert(n);
	
	printf("this: ");
	printwalllist(n->wl);
	
	if (n->front) {
		lev++;
		for (i=0; i<lev; i++) putchar(' ');
		for (i=0; i<lev; i++) putchar(' ');
		printf("front: ");
		preorderprint(n->front);
		lev--;
	}
	if (n->back) {
		lev++;
		for (i=0; i<lev; i++) putchar(' ');
		for (i=0; i<lev; i++) putchar(' ');
		printf("back: ");
		preorderprint(n->back);
		lev--;
	}
}

void printwalllist(wlelem *wl) {
	wlelem *we;
	int n;
	
	n = 0;
	we = wl->next;

	while(we) {
		printwall(&we->w);
		n++;
		we = we->next;
	}
	printf("%d walls\n", n);
}

void printwall(wall *w) {
	printf("%lf %lf -> %lf %lf\n", w->start[0], w->start[1],
			w->end[0], w->end[1]);
	printf("normal: %lf %lf\n", w->n[0], w->n[1]);
}

void bresenham(int x1, int y1, int x2, int y2, byte c) {
	int dx, dy, x, y, xinc, error;
	int i;
	
	if (y2 < y1) {
		y = y2;
		y2 = y1;
		y1 = y;
		x = x2;
		x2 = x1;
		x1 = x;
	}

	if (x2 > x1) xinc = 1;
	else if (x2 < x1) xinc = -1;
	else {
		for (y = y1; y <= y2; y++) {
			plot(x1, y, c);
		}
		return;
	}
	
	dx = x2 - x1;
	dy = y2 - y1;
	if (dx < 0) dx = -dx;
	error = -dx / 2;
	x = x1;
	y = y1;

	if (!dy) {
		for (i=0; i<=dx; i++) {
			plot(x, y, c);
			x += xinc;
		}
		return;
	}

	for (i=0; i<=dx; i++) {
		plot(x, y, c);
		x += xinc;
		error += dy;
		while (error > 0 && y < y2) {
			y++;
			plot(x, y, c);
			error -= dx;
		}
	}

}

void plot(int x, int y, byte c) {
	if (x < 0 || x >= sc.swidth || y < 0 || y >= sc.sheight) return;
	img[y * sc.swidth + x] = c;
}

void loadtex(int n, char *filename) {
	int i;

	if (!(infile = fopen(filename, "r"))) {
			perror("fopen");
			exit(1);
	}

	readin("%d", &tex[n].width, "texture width");
	readin("%d", &tex[n].height, "texture height");
	tex[n].wmask = tex[n].width - 1;
	tex[n].hmask = tex[n].height - 1;
	for (i=0; i<8; i++) {
		if (tex[n].height & (1 << i)) {
			tex[n].logheight = i;
			break;
		}
	}

	tex[n].data = (byte *) xmalloc(tex[n].width * tex[n].height);
	for (i=0; i<tex[n].width * tex[n].height; i++) {
		readin("%d", &tex[n].data[i], "texture pixel #%d", i);
	}

	fclose(infile);
}

void showtexture(int x, int y, texture_map *t) {
	int i;
	int j;

	for (i=0; i<t->height; i++) {
		for (j=0; j<t->width; j++) {
			img[(y + i) * sc.swidth + x + j] = t->data[i * t->height + j];
		}
	}

	drawimage(img);
}
