/* Polynomial surface routines
 * Ben Lin	started 3/9/98 (only polynomial stuff)
 * 4/9/98 - added parser, surface routines
 */

/* Can't handle ^0 properly */

#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>

#define samesign(a, b) (((a) * (b)) > 0)

enum {
	maxdegree = 20,
	maxstring = 100,
	ascii_count = 128,
	t_white = 0,
	t_digit,
	t_dpoint,
	t_exp,
	t_x,
	t_y,
	t_z,
	t_power, t_times, t_plus, t_minus, t_lparen, t_rparen,
	t_semicolon,
	t_integer, t_double,
	t_unknown,
	WARNING,
	FATAL
};

typedef struct {
	int n;
	int d;
} rational;

typedef double vector[3];

// Multinomial term
typedef struct multiterm {
	vector deg;
	double coeff;
	struct multiterm *next;
} multiterm;

typedef struct surftree {
	struct surftree *left, *right;
	char type;
	union {
		double value;
		int power;
	} u;
} surftree;

//typedef rational polynomial[maxdegree];
typedef struct {
	double c[maxdegree];
	int deg;
} polynomial;

//Prototypes
void findroot(void);
double bisection(double, double);
double evalpoly(double);
void readsurface(void);
void initchtable(char *, char);
void gettoken(void);
void errmsg(char, char *, ...);
void printsurf(surftree *);
void parsetop(surftree *);
void parsetimes(surftree *);
char isterm(char);
void parsepower(surftree *);
void parseprimitive(surftree *);
void parseinteger(surftree *);

double evalsurf(surftree *, vector);
void derive(surftree *, surftree *, char);
void copysurf(surftree *, surftree *);

void makepoly(polynomial *, surftree *, vector);
void polypoly(polynomial *, polynomial *, polynomial *);
void addpoly(polynomial *, polynomial *, polynomial *);
void subpoly(polynomial *, polynomial *, polynomial *);
void fprintpoly(FILE *, polynomial *);
void printpoly(polynomial *);
void copypoly(polynomial *, polynomial *);

void *salloc(size_t);

double tolerance = 1e-8;

polynomial poly;

//multiterm *surf;
//multiterm *surfnorm[3];
surftree *surf = NULL;
surftree *normal[3];
vector transsurf[3];
char chtable[ascii_count];

struct {
	char type;
	union {
		int power;
		double value;
	} u;
} token;

int main(int argc, char **argv) {
	int i;
	vector v;

	/*
	p[0].n = 6;
	p[1].n = 5;
	p[2].n = 1;
	p[0].d = 1;
	p[1].d = 1;
	p[2].d = 1;
	*/
	readsurface();
	v[0] = v[1] = v[2] = 1;
	printsurf(surf);
	printf("\n");
	printf("at 1,1,1 it's %f\n", evalsurf(surf, v));

	for (i=0; i<3; i++) {
		normal[i] = (surftree *) salloc(sizeof(surftree));
		derive(surf, normal[i], i);
		printf("Normal %d: ", i);
		printsurf(normal[i]);
		printf("\n");
	}

	if (0) {
		poly.c[2] = 1;
		poly.c[1] = 0;
		poly.c[0] = -169;
		poly.deg = 2;
		printpoly(&poly);
	}
	makepoly(&poly, surf, v);
	printpoly(&poly);

	findroot();

}

void findroot(void) {
	double d, e, f, g;

	f = evalpoly(0);
	d = 0;

	if (f > 0) {
		do {
			e = d;
			d += 0.01;
			g = evalpoly(d);
		} while (d<100 && g >= 0);
		f = bisection(e, d);
	} else if (f < 0) {
		do {
			e = d;
			d += 0.01;
			g = evalpoly(d);
		} while (d<100 && g <= 0);
		f = bisection(e, d);
	}

	printf("%f is a root\n", f);
}

double bisection(double min, double max) {
	double d, e, f, g;
	char flag;

	flag = 0;

	f = evalpoly(min);
	g = evalpoly(max);
	if (samesign(f, g)) {
		return(0);
	}

	while (1) {
		if (fabs(f) <= tolerance) break;
		if (fabs(g) <= tolerance) {
			min = max;
			break;
		}
		else {
			//printf("%f %f\n", f, g);
			d = (min + max) * 0.5;
			e = evalpoly(d);
			if (samesign(e, f)) {
				min = d;
				f = e;
			} else {
				max = d;
				g = e;
			}
			//printf("%lf\n", d);
		}
	}

	return(min);
}

double evalpoly(double x) {
	double result;
	int i;

	//Horner's rule
	i = poly.deg;
	result = poly.c[i];
	do {
		result *= x;
		i--;
		result += poly.c[i];
	} while (i>0);

	return(result);
}

void initchtable(char *s, char n) {
	int i;

	i = 0;
	while(s[i]) {
		chtable[s[i]] = n;
		i++;
	}
}

// Based on Jeff Kingston's tt parser
void readsurface(void) {
	int i;

	for (i=0; i<ascii_count; i++) chtable[i] = t_unknown;
	initchtable("\t\n ", t_white);
	initchtable("0123456789", t_digit);
	initchtable("x", t_x);
	initchtable("y", t_y);
	initchtable("z", t_z);
	initchtable("eE", t_exp);
	initchtable("^", t_power);
	initchtable("*", t_times);
	initchtable("+", t_plus);
	initchtable("-", t_minus);
	initchtable("(", t_lparen);
	initchtable(")", t_rparen);
	initchtable(".", t_dpoint);
	initchtable(";", t_semicolon);

	surf = (surftree *) salloc(sizeof(surftree));
	gettoken();
	parsetop(surf);
	if (token.type != t_semicolon) {
		errmsg(WARNING, "missing ';' after polynomial");
	}
}

// Get next token (puts result in the token struct)
void gettoken(void) {
	char ch;
	char s[maxstring];
	double d;
	int i;

	while (chtable[ch = getchar()] == t_white);
	
	// Spaghetti: need this for the case part to work
	if (chtable[ch] == t_dpoint) {
		token.u.power = 0;
	}

	switch(chtable[ch]) {
		case t_digit:
			token.type = t_integer;
			token.u.power = ch - '0';
			while (chtable[ch = getchar()] == t_digit) {
				token.u.power *= 10;
				token.u.power += ch - '0';
			}


		case t_dpoint:
			if (chtable[ch] == t_dpoint) {
				token.u.value = token.u.power;
				token.type = t_double;
				ch = getchar();
				if (chtable[ch] != t_digit) {
					errmsg(WARNING, "decimal point with no digit after it");
				}
				d = 0.1;
				while (chtable[ch] == t_digit) {
					token.u.value += d * (ch - '0');
					d *= 0.1;
					ch = getchar();
				}
			}

			if (chtable[ch] == t_exp) {
				ch = getchar();
				if (chtable[ch] != t_digit) {
					errmsg(FATAL, "empty exponent");
				}
				if (token.type != t_double) {
					errmsg(FATAL, "cannot use exponent on power");
				} else {
					i = ch - '0';
					while (chtable[ch = getchar()] == t_digit) {
						i *= 10;
						i += ch - '0';
					}
					do {
						token.u.value *= 10;
						i--;
					} while (i > 0);
				}

			}
			ungetc(ch, stdin);
			break;

		default:
			token.type = chtable[ch];
			break;

	}
}

void parsetop(surftree *root) {

	root->left = (surftree *) salloc(sizeof(surftree));

	parsetimes(root->left);

	if (token.type == t_plus || token.type == t_minus) {
		root->type = token.type;
		gettoken();

		root->right = (surftree *) salloc(sizeof(surftree));
		parsetop(root->right);
	} else {
		*root = *(root->left);
	}
}

void parsetimes(surftree *root) {

	root->left = (surftree *) salloc(sizeof(surftree));

	parsepower(root->left);

	if (token.type == t_times || isterm(token.type)) {
		root->type = t_times;
		if (token.type == t_times) gettoken();

		root->right = (surftree *) salloc(sizeof(surftree));
		parsetimes(root->right);
	} else {
		*root = *(root->left);
	}
}

void parsepower(surftree *root) {

	root->left = (surftree *) salloc(sizeof(surftree));

	parseprimitive(root->left);

	if (token.type == t_power) {
		root->type = t_power;
		gettoken();

		root->right = (surftree *) salloc(sizeof(surftree));
		parseinteger(root->right);
	} else {
		*root = *(root->left);
	}
}

void parseprimitive(surftree *root) {
	if (token.type == t_minus) {
		/* unary minus not very efficient */
		/* simulated by 0 - x */
		gettoken();
		root->left = (surftree *) salloc(sizeof(surftree));
		root->right = (surftree *) salloc(sizeof(surftree));
		root->left->type = t_double;
		root->left->left = root->left->right = NULL;
		root->left->u.value = 0;
		root->type = t_minus;
		parseprimitive(root->right);
		return;
	}

	switch (token.type) {
		case t_double:
			root->left = root->right = NULL;
			root->type = t_double;
			root->u.value = token.u.value;
			break;

		case t_integer:
			root->left = root->right = NULL;
			// convert integers to doubles
			root->type = t_double;
			root->u.value = (double) token.u.power;
			break;

		case t_lparen:
			gettoken();
			parsetop(root);
			if (token.type != t_rparen) {
				errmsg(WARNING, "missing ')'");
			}
			break;

		case t_x:
		case t_y:
		case t_z:
			root->left = root->right = NULL;
			root->type = token.type;
			break;

		default:
			errmsg(FATAL, "primitive expected");
			break;
	}
	gettoken();
}

void parseinteger(surftree *root) {
	if (token.type != t_integer) {
		errmsg(FATAL, "integer expected");
	}

	root->left = root->right = NULL;
	root->type = t_integer;
	root->u.power = token.u.power;
	gettoken();
}


// is the token the beginning of a term?
// (needed so that I can use juxtaposition as multiplication)
char isterm(char n) {
	switch (n) {
		case t_x:
		case t_y:
		case t_z:
		case t_integer:
		case t_double:
		case t_lparen:
			return(1);
			break;
		default:
			return(0);
			break;
	}
}

void errmsg(char type, char *format, ...) {
	va_list ap;

	switch (type) {
		case WARNING:
			fprintf(stderr, "Warning: ");
			break;

		case FATAL:
			fprintf(stderr, "Fatal error: ");
			break;
	}

	va_start(ap, format);
	fprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, "\n");

	if (type == FATAL) exit(3);
}

void *salloc(size_t size) {
	void *ptr;

	if (!(ptr = (void *) malloc(size))) {
		errmsg(FATAL, "Out of memory!");
	}

	return(ptr);
}

void printsurf(surftree *root) {
	if ((root->type >= t_power) && (root->type <= t_minus)) {
		switch(root->type) {
			case t_power:
				printf("^");
				break;

			case t_times:
				printf("*");
				break;

			case t_minus:
				printf("-");
				break;

			case t_plus:
				printf("+");
				break;
		}

		printf("(");
		if (!root->left) {
			errmsg(FATAL, "buggy tree missing left side!");
		}
		printsurf(root->left);
		if (!root->right) {
			errmsg(FATAL, "buggy tree missing right side!");
		}
		printf(", ");
		printsurf(root->right);
		printf(")");

	} else {
		switch (root->type) {
			case t_double:
				printf("%f", root->u.value);
				break;

			case t_integer:
				printf("%d", root->u.power);
				break;

			case t_x:
				printf("x");
				break;

			case t_y:
				printf("y");
				break;

			case t_z:
				printf("z");
				break;
		}
		if (root->left || root->right) {
			errmsg(FATAL, "buggy tree has extra stuff!");
		}
	}
}

double evalsurf(surftree *root, vector v) {
	double f, r;
	int p;

	switch (root->type) {
		case t_x:
			return(v[0]);
			break;

		case t_y:
			return(v[1]);
			break;

		case t_z:
			return(v[2]);
			break;

		case t_double:
			return(root->u.value);
			break;

		case t_integer:
			return(root->u.power);
			break;

		case t_plus:
			return(evalsurf(root->left, v) + evalsurf(root->right, v));
			break;

		case t_minus:
			return(evalsurf(root->left, v) - evalsurf(root->right, v));
			break;

		case t_times:
			return(evalsurf(root->left, v) * evalsurf(root->right, v));
			break;

		case t_power:
			r = f = evalsurf(root->left, v);
			p = root->right->u.power;
			while (p>1) {
				r *= f;
				p--;
			}
			return(r);
			break;
			
		default:
			errmsg(FATAL, "Can't handle this tree!");
			break;
	}
}

// Symbolic differentiation
void derive(surftree *src, surftree *dst, char n) {
	// dx/dx = dy/dy = dz/dz = 1
	if (src->type == (t_x + n)) {
		dst->left = dst->right = NULL;
		dst->type = t_double;
		dst->u.value = 1;
		return;
	}

	switch (src->type) {
		// dc/dx = 0, where c = constant or other variable
		case t_double:
		case t_integer:
		case t_x:
		case t_y:
		case t_z:
			dst->left = dst->right = NULL;
			dst->type = t_double;
			dst->u.value = 0;
			break;

		// d(f +- g)/dx = df/dx +- dg/dx
		case t_plus:
		case t_minus:
			dst->type = src->type;
			dst->left = (surftree *) salloc(sizeof(surftree));
			dst->right = (surftree *) salloc(sizeof(surftree));
			derive(src->left, dst->left, n);
			derive(src->right, dst->right, n);
			break;

		// product rule
		// d(fg)/dx = df/dx * g + dg/dx * f
		case t_times:
			dst->type = t_plus;
			dst->left = (surftree *) salloc(sizeof(surftree));
			dst->right = (surftree *) salloc(sizeof(surftree));

			dst->left->type = t_times;
			dst->left->left = (surftree *) salloc(sizeof(surftree));
			dst->left->right = (surftree *) salloc(sizeof(surftree));
			derive(src->left, dst->left->left, n);
			copysurf(src->right, dst->left->right);
			
			dst->right->type = t_times;
			dst->right->left = (surftree *) salloc(sizeof(surftree));
			dst->right->right = (surftree *) salloc(sizeof(surftree));
			derive(src->right, dst->right->left, n);
			copysurf(src->left, dst->right->right);
			break;

		// chain rule
		// d(f^n)/dx = n * df/dx * f^(n - 1)
		case t_power:
			// special case
			if (src->right->u.power == 1) {
				derive(src->left, dst, n);
				return;
			}
			dst->type = t_times;
			dst->left = (surftree *) salloc(sizeof(surftree));
			dst->right = (surftree *) salloc(sizeof(surftree));

			// `n *'
			dst->left->type = t_double;
			dst->left->u.value = src->right->u.power;
			dst->left->left = dst->left->right = NULL;

			// right subtree is `df/dx * f^(n - 1)'
			dst->right->type = t_times;
			dst->right->left = (surftree *) salloc(sizeof(surftree));
			dst->right->right = (surftree *) salloc(sizeof(surftree));
			derive(src->left, dst->right->left, n);

			// `f^(n - 1)'
			copysurf(src, dst->right->right);
			dst->right->right->right->u.power--;
			break;
	}
}

void copysurf(surftree *src, surftree *dst) {
	// TODO: I should only copy type and u
	*dst = *src;
	if (src->left) {
		dst->left = (surftree *) salloc(sizeof(surftree));
		copysurf(src->left, dst->left);
	} else {
		dst->left = NULL;
	}

	if (src->right) {
		dst->right = (surftree *) salloc(sizeof(surftree));
		copysurf(src->right, dst->right);
	} else {
		dst->right = NULL;
	}
}

void makepoly(polynomial *p, surftree *root, vector v) {
	int i;
	polynomial q, r;

	switch (root->type) {
		case t_double:
			p->deg = 0;
			p->c[0] = root->u.value;
			break;

		case t_x:
		case t_y:
		case t_z:
			p->deg = 1;
			p->c[0] = 0;
			p->c[1] = v[root->type - t_x];
			break;

		case t_minus:
			makepoly(&q, root->left, v);
			makepoly(&r, root->right, v);
			subpoly(p, &q, &r);
			break;
			
		case t_plus:
			makepoly(&q, root->left, v);
			makepoly(&r, root->right, v);
			addpoly(p, &q, &r);
			break;
			
		case t_times:
			makepoly(&q, root->left, v);
			makepoly(&r, root->right, v);
			polypoly(p, &q, &r);
			break;

		case t_power:
			makepoly(p, root->left, v);
			copypoly(p, &q);
			for (i=1; i<root->right->u.power; i++) {
				polypoly(&r, p, &q);
				copypoly(&r, p);
			}
			break;
	}
}

// sets p = q * r
void polypoly(polynomial *p, polynomial *q, polynomial *r) {
	int i, j;

	p->deg = q->deg + r->deg;
	if (p->deg > maxdegree) {
		errmsg(FATAL, "degree of polynomial too large");
	}

	for (i=0; i<=p->deg; i++) {
		p->c[i] = 0;
	}

	for (i=0; i<=r->deg; i++) {
		for (j=0; j<=q->deg; j++) {
			p->c[i + j] += r->c[i] * q->c[j];
		}
	}
}

// sets p = q + r
void addpoly(polynomial *p, polynomial *q, polynomial *r) {
	int i;

	if (q->deg > r->deg) p->deg = q->deg;
	else p->deg = r->deg;

	for (i=p->deg; i>=0; i--) {
		p->c[i] = q->c[i] + r->c[i];
		if (p->c[p->deg] == 0) p->deg--;
	}

	if (p->deg < 0) p->deg = 0;
}

// sets p = q - r
void subpoly(polynomial *p, polynomial *q, polynomial *r) {
	int i;

	if (q->deg > r->deg) p->deg = q->deg;
	else p->deg = r->deg;

	for (i=p->deg; i>=0; i--) {
		p->c[i] = q->c[i] - r->c[i];
		if (p->c[p->deg] == 0) p->deg--;
	}

	if (p->deg < 0) p->deg = 0;
}

void copypoly(polynomial *src, polynomial *dst) {
	int i;

	for (i=0; i<=src->deg; i++) {
		dst->c[i] = src->c[i];
	}
	dst->deg = src->deg;
}

void printpoly(polynomial *p) {
	fprintpoly(stdout, p);
}

void fprintpoly(FILE *file, polynomial *p) {
	int i;
	char first;

	first = 0;
	for (i=p->deg; i>0; i--) {
		if (fabs(p->c[i]) > tolerance) {
			if (first && p->c[i] > 0) fprintf(file, "+");
			fprintf(file, "%lfx", p->c[i]);
			if (i > 1) fprintf(file, "^%d", i);
			first = 1;
		}
	}

	if (!first) {
		fprintf(file, "%lf", p->c[0]);
	} else if (p->c[0]) {
		fprintf(file, "%+lf", p[0]);
	}
	fprintf(file, "\n");
}

