/* Ray tracer
 * 25/8/98
   Ben Lin
   
   27/8/98 - it works! shading not very good though
   28/8/98 - transformed the lights, optimized
   29/8/98 - a few bug fixes; I shouldn't have scaled the normals,
   			and I had been using abs instead of fabs!

   // fix up transscene - quite roundabout currently
   */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graphapp.h"

#define normvec(a) dotvec(a, a)
#define lenvec(a) sqrt(dotvec(a, a))

#define D0 1	// For use in diffuse reflection formula

typedef double vector[3];

struct sceneinfo {
	vector eye;
	vector at;
	vector up;
	double focal;
	double wwidth;
	double wheight;
	int swidth;
	int sheight;
};

struct facestruct {
	int ecount;
	int (*e)[2];
	double n[3];	//	normal
	double sn[3];	// normal after transformation
	double colour[3];
	double diffuse, spec;
	int specpower;

	//	extra info:
	//struct objstruct *o;	// owner
	char hidden;	//self-hidden

	int min[3];	// bounding box
	int max[3];
	int maxn;	// max component of the normal - used in inface
};

struct objstruct {
	int vcount;			// no. of vertices
	//double (*v)[3];		// vertices
	int fcount;			// no. of faces
	struct facestruct *face;
};

struct ilist {
	struct ilist *next;
	int i;
};

struct facelist {
	struct facelist *next;
	struct facestruct *face;
};

struct lightstruct {
	double loc[3];	// location
	double sloc[3];	// location after transformation
	double colour[3];
};

void readin(char *, void *, char *);
void parsecomment(void);
void scanfill(struct facestruct *);
//void drawscene(drawing, rect);
void drawscene(void);
void loadfile(char *);
void transscene(void);
void render(void);
void raytrace(double *, int);
double findinter(struct facestruct *, double *);
void outline(struct facestruct *);
void *salloc(size_t);
void wireframe(void);
void setcolourpt(double *, struct facestruct *);

void showvec(double *);
void addvec(double *, double *, double *);
void subvec(double *, double *, double *);
void multvec(double *, double, double *);
void matvec(double *, double [3][3], double *);
void matmat(double [3][3], double [3][3], double [3][3]);
void assignmat(double [3][3], double [3][3]);
void showmat(double [3][3]);
double dotvec(double *, double *);

char inface(double [3], struct facestruct *);

void initfacelist(void);

double amblight[3];
int lightcount;
int objcount;
int vcount;
struct sceneinfo s;
struct objstruct *o;
struct lightstruct *light;
double (*vertex)[3];
double (*svertex)[3];	// vertex coords after translation and rotation
int (*scoord)[2];	// screen coords after projection
double seye;		// eye z coord after transformation; at and up are trivial

struct facelist *active, *passive;

FILE *infile;

int main(int argc, char **argv) {

	/*
	window w;
	drawing d;
	*/

	startlog();
	loadfile("tt");
	transscene();

#if 0
	w = newwindow("Ray", rect(50,50,s.swidth,s.sheight), StandardWindow);
	d = newdrawing(rect(0,0,s.swidth,s.sheight), drawscene);
	setdata(w, d);	/* store the pointer with the window */

	show(w);
	mainloop();
#endif
	getchar();
	drawscene();
	getchar();

	return(0);
}

/* Ray trace */

void render(void) {
	int i, j;
	double ray[3], v[3];
	double colour[3];
	struct facelist *fp, *temp;

	ray[0] = ray[1] = 0;
	ray[2] = seye;

	if (0) {
	// Eliminate self-hidden faces
	// Assume all normals point out
	// Doesn't seem to work
	for (i=0; i<objcount; i++) {
		if (o[i].fcount < 3) {
			// Degenerates don't count
			for (j=0; j<o[i].fcount; j++) {
				o[i].face[j].hidden = 0;
				outline(&o[i].face[j]);
			}
		} else {
			for (j=0; j<o[i].fcount; j++) {
				subvec(v, ray, svertex[o[i].face[j].e[0][0]]);
				if (dotvec(v, o[i].face[j].sn) < 0) {
					o[i].face[j].hidden = 1;
					outline(&o[i].face[j]);
					//printf("hidden\n");
				} else {
					o[i].face[j].hidden = 0;
					outline(&o[i].face[j]);
					//printf("visible\n");

				}

			}
		}
	}
	
	}

	ray[1] = -s.sheight / 2;
	ray[2] = seye;

	passive = active->next;
	for (j=0; j<s.sheight; j++, ray[1]++) {
		while (passive && (passive->face->min[1] <= j)) {
			passive = passive->next;
		}
		fp = active;
		while (fp->next != passive) {
			if (fp->next->face->max[1] < j) {
				temp = fp->next->next;
				free(fp->next);
				fp->next = temp;
			} else {
				fp = fp->next;
			}
		}
		/*
		logprintf("\n\nactive:");
		fp = active;
		while (fp->next) {
			fp = fp->next;
			logprintf("%d\t", fp->face->min[1]);
		}
			logprintf("\n\nactive to passive %d:", j);
		fp = active->next;
		while (fp != passive) {
			logprintf("%d\t", fp->face->min[1]);
			fp = fp->next;
		}
		*/

		ray[0] = -s.swidth / 2;
		for (i=0; i<s.swidth; i++, ray[0]++) {
			setcolour(Yellow);
			drawpoint(pt(i, j));
			raytrace(ray, i);
			drawpoint(pt(i, j));
		}
	}
}

void raytrace(double *ray, int x) {
	int i, j;
	double mu, minmu;
	double d[3];
	double v[3];
	struct facestruct *face, *minface;
	struct facelist *fp;

	multvec(d, 1 / lenvec(ray), ray);
	/*
	logprintf("ray (l = %lf):", lenvec(ray));
	logvec(ray);
	logprintf("d:");
	logvec(d);
	*/

	minmu = -1;
	fp = active->next;
	while (fp != passive) {
		if (x >= fp->face->min[0] && x <= fp->face->max[0]) {
			mu = findinter(fp->face, d);

			logprintf("mu: %lf\n", mu);
			if (mu > 0) {
				multvec(v, mu, d);

				/* Should be 0
				subvec(v2, v, svertex[o[i].face[j].e[0][0]]);
				printf("%lf\n", dotvec(v2, o[i].face[j].n));
				*/

				// If ray is in the face, record it
				if (inface(v, fp->face)) {
					// update minmu if necessary
					if (minmu == -1) {
						minmu = mu;
						minface = fp->face;
					} else if (minmu > mu) {
						minmu = mu;
						minface = fp->face;
					}
				}
			}
		}
		fp = fp->next;
	}

	if (minmu > 0) {
		// work out colour of point
		multvec(v, minmu, d);

		setcolourpt(v, minface);
	} else {
		setcolour(Black);
	}
}

// Find pt of intersection between line and plane
double findinter(struct facestruct *f, double *d) {
	double denom;

	// mu = (s - n . P1) / (n . d)
	// where the line is P = P1 + mu . d
	// and the plane is n . P = s
	denom = dotvec(f->sn, d);
	if (!denom) return(-1);
	return(dotvec(f->sn, svertex[f->e[0][0]]) / denom);
}

/* Parse a .gen file */

void loadfile(char *filename) {
	int i, j, k;
	int v;
	int n;
	double f;
	int *fromv, *tov;

	if (!(infile = fopen(filename, "r"))) {
			perror("fopen");
			exit(1);
	}

	readin("%lf", &s.eye[0], "eye x");
	readin("%lf", &s.eye[1], "eye y");
	readin("%lf", &s.eye[2], "eye z");
	readin("%lf", &s.at[0], "view at x");
	readin("%lf", &s.at[1], "view at y");
	readin("%lf", &s.at[2], "view at z");
	readin("%lf", &s.up[0], "view up x");
	readin("%lf", &s.up[1], "view up y");
	readin("%lf", &s.up[2], "view up z");
	readin("%lf", &s.focal, "focal length");
	readin("%lf", &s.wwidth, "window width");
	readin("%lf", &s.wheight, "window height");
	readin("%d", &s.swidth, "screen height");
	readin("%d", &s.sheight, "screen height");

	readin("%lf", &amblight[0], "ambient light r");
	readin("%lf", &amblight[1], "ambient light g");
	readin("%lf", &amblight[2], "ambient light b");

	readin("%d", &lightcount, "number of lights");

	light = (struct lightstruct *) salloc(sizeof(struct lightstruct) * lightcount);

	for (i=0; i<lightcount; i++) {
		readin("%lf", &light[i].loc[0], "light x");
		readin("%lf", &light[i].loc[1], "light y");
		readin("%lf", &light[i].loc[2], "light z");

		readin("%lf", &light[i].colour[0], "light r");
		readin("%lf", &light[i].colour[1], "light g");
		readin("%lf", &light[i].colour[2], "light b");
	}

	readin("%d", &objcount, "number of objects");
	o = (struct objstruct *) salloc(sizeof(struct objstruct) * objcount);

	readin("%d", &vcount, "total vertices");
	vertex = (double (*)[3]) salloc(sizeof(double) * 3 * vcount);
	svertex = (double (*)[3]) salloc(sizeof(double) * 3 * vcount);
	scoord = (int (*)[2]) salloc(sizeof(int) * 2 * vcount);
	
	fromv = (int *) salloc(sizeof(int) * vcount);
	tov = (int *) salloc(sizeof(int) * vcount);

	v = 0;

	for (i=0; i<objcount; i++) {

		readin("%d", &o[i].vcount, "number of vertices in object");

		//o[i].v = (double (*)[3]) salloc(sizeof(double) * 3 * o[i].vcount);
		for (j=0; j<o[i].vcount; j++) {
			readin("%lf", &vertex[v][0], "vertex x");
			readin("%lf", &vertex[v][1], "vertex y");
			readin("%lf", &vertex[v][2], "vertex z");
			v++;
		}

		readin("%d", &o[i].fcount, "number of faces");
		o[i].face = (struct facestruct *) salloc(sizeof(struct facestruct) * o[i].fcount);
		
		printf("object %d\n", i);
		for (j=0; j<o[i].fcount; j++) {
			//printf("\tface %d\n", j);

			k = 0;
			do {
				readin("%d", &n, "edge's from-vertex");

				if (n != -1) {
					fromv[k] = v - o[i].vcount + n;
					readin("%d", &n, "edge's to-vertex");
					tov[k] = v - o[i].vcount + n;
					k++;
				}
			} while (n != -1);
			o[i].face[j].ecount = k;
			o[i].face[j].e = (int (*)[2]) salloc(sizeof(int) * 2 * o[i].face[j].ecount);

			/* Renumber consistently */
			o[i].face[j].e[0][0] = fromv[0];
			o[i].face[j].e[0][1] = tov[0];
			fromv[0] = tov[0] = -1;
			for (k=1; k<o[i].face[j].ecount; k++) {
				n = 1;
				while (1) {
					if (fromv[n] == o[i].face[j].e[k - 1][1]) {
						o[i].face[j].e[k][0] = fromv[n];
						o[i].face[j].e[k][1] = tov[n];
						fromv[k] = tov[k] = -1;
						break;
					} else if (tov[n] == o[i].face[j].e[k - 1][1]) {
						o[i].face[j].e[k][0] = tov[n];
						o[i].face[j].e[k][1] = fromv[n];
						fromv[k] = tov[k] = -1;
						break;
					}
				 	n++;
					if (n > o[i].face[j].ecount) {
						fprintf(stderr, "face not closed\n");
						exit(1);
					}
				}

			}

			readin("%lf", &o[i].face[j].n[0], "normal x");
			readin("%lf", &o[i].face[j].n[1], "normal y");
			readin("%lf", &o[i].face[j].n[2], "normal z");

			/* Colour */
			readin("%lf", &o[i].face[j].colour[0], "face r");
			readin("%lf", &o[i].face[j].colour[1], "face g");
			readin("%lf", &o[i].face[j].colour[2], "face b");

			/* Diffuse, spec(ular), spec. power */
			readin("%lf", &o[i].face[j].diffuse, "diffuse");
			readin("%lf", &o[i].face[j].spec, "specular");
			readin("%d", &o[i].face[j].specpower, "specular power");

		}

	}
	
	free(fromv);
	free(tov);

}

//void drawscene(drawing d, rect r) {
void drawscene(void) {
	int i, j;

	init();
	setcolour(Black);
	fillrect(initrect(0, 0, s.swidth, s.sheight));

	initfacelist();
	wireframe();
	render();

	/*
	for (int n=0; n<o[i].face[j].ecount; n++) {
		int a, b;
		
		a = o[i].face[j].e[n][0];
		b = o[i].face[j].e[n][1];
		printf("%d -> %d: \n", a, b);
		printf("%d, %d -> %d, %d\n", scoord[a][0], scoord[a][1],
				scoord[b][0], scoord[b][1]);
	}

	outline(&o[i].face[j]);
	scanfill(&o[i].face[j]);
	*/

	/*
	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			//printf("painting %d %d\n", i, j);
			scanfill(&o[i].face[j]);
		}
	}
	*/
}

void scanfill(struct facestruct *f) {
	int i, j;
	int ymin;
	int max;
	struct ilist *inter, *ip, *temp;
	int error;
	int x, y, dx, dy, x2, y2;
	int endx, endy;

	setcolour(initrgb((int) (f->colour[0]*255), (int) (f->colour[1]*255), (int) (f->colour[2]*255)));

	ymin = f->min[1];

	max = f->max[1] - ymin + 1;
	inter = (struct ilist *) salloc(sizeof(struct ilist) * max);
	for (i=0; i<max; i++) {
		inter[i].next = NULL;
	}

	//printf("%d %d\n", ymin, f->max[1]);

	/* Calculate intersections */

	for (i=0; i<f->ecount; i++) {
		x = scoord[f->e[i][0]][0];
		y = scoord[f->e[i][0]][1];
		endx = scoord[f->e[i][1]][0];
		endy = scoord[f->e[i][1]][1];
		//drawline(pt(x, y), pt(endx, endy));
		//printf("%d %d, %d %d\n", x, y, endx, endy);
		dx = endx - x;
		dy = endy - y;
		if (dx > 0) x2 = 1;
		else {
			dx = -dx;
			x2 = -1;
		}
		if (dy > 0) {
			y2 = 1;
		} else {
			dy = -dy;
			y2 = -1;
		}

		if (dy) {
			error = dy;
			if (error > 0) error = -error;
			error /= 2;
			y-=ymin;
			endy-=ymin;
			do {

				if (!inter[y].next) {
					inter[y].next = (struct ilist *) salloc(sizeof(struct ilist));
					inter[y].next->i = x;
					inter[y].next->next = NULL;
				} else {
					ip = &inter[y];
					while (ip->next && ip->next->i < x)
						ip = ip->next;

					temp = ip->next;
					ip->next = (struct ilist *) salloc(sizeof(struct ilist));
					ip->next->i = x;
					ip->next->next = temp;
				}

				y += y2;

				error += dx;
				while (error > 0) {
					x += x2;
					error -= dy;
				}
			} while(y != endy);

			/* Must decide whether to include this vertex or not */
			if (i == f->ecount - 1) {
				dy = scoord[f->e[0][1]][1] - scoord[f->e[0][0]][1];
			} else {
				dy = scoord[f->e[i+1][1]][1] - scoord[f->e[i+1][0]][1];
			}
			
			if (dy > 0) dy = 1;
			else if (dy < 0) dy = -1;
			else dy = -1;
			
			if (y2 != dy || dy == 0) {

				/* More efficient if I duplicate code ? */
				if (!inter[y].next) {
					inter[y].next = (struct ilist *) salloc(sizeof(struct ilist));
					inter[y].next->i = x;
					inter[y].next->next = NULL;
				} else {
					ip = &inter[y];
					while (ip->next && ip->next->i < x)
						ip = ip->next;

					temp = ip->next;
					ip->next = (struct ilist *) salloc(sizeof(struct ilist));
					ip->next->i = x;
					ip->next->next = temp;
				}
			}

		}
	}

	/* Draw the scan lines */
	for(i=0; i<max; i++) {
		ip = &inter[i];
		while (ip->next) {
			ip = ip->next;
			x = ip->i;
			if (!ip->next) continue;
			ip = ip->next;
			x2 = ip->i;
			//printf("%d: %d %d\n", i, x, x2);
			drawline(pt(x, i + ymin), pt(x2, i + ymin));
		}
	}

	/* Free memory */
	for(i=0; i<max; i++) {
		ip = inter[i].next;
		while (ip) {
			temp = ip->next;
			free(ip);
			ip = temp;
		}
	}
	free(inter);
}

void outline(struct facestruct *f) {
	int i, j;
	int x1, y1, x2, y2;

	setcolour(initrgb((int) (f->colour[0]*255), (int) (f->colour[1]*255), (int) (f->colour[2]*255)));
	for (i=0; i<f->ecount; i++) {
		x1 = scoord[f->e[i][0]][0];
		y1 = scoord[f->e[i][0]][1];
		x2 = scoord[f->e[i][1]][0];
		y2 = scoord[f->e[i][1]][1];
		//printf("%d %d, %d %d\n", x1, y1, x2, y2);
		drawline(pt(x1, y1), pt(x2, y2));

	}
}

void readin(char *s, void *f, char *thing) {
	char done;
	int check;
	char ch;

	done = 0;
	do {
		fscanf(infile, " %c", &ch);
		if (ch == '/') {
			parsecomment();
		} else {
			ungetc(ch, infile);
			check = fscanf(infile, s, f);
			if (!check) {
				fprintf(stderr, "Error reading %s\n", thing);
				exit(1);
			}
			//printf("Read %s\n", thing);
			done++;
		}
	} while (!done);
}

void parsecomment(void) {
	char ch;
	char flag;

	if ((ch = getc(infile)) != '*') {
		fprintf(stderr, "Warning: bad comment\n");
	}
	flag = 0;
	do {
		do {
			ch = getc(infile);
		} while (ch != '*');
		if ((ch = getc(infile)) == '/') {
			flag++;
		}
	} while (!flag);
}

// Transform the coords of each vertex so that the camera faces the z=0 plane
/* This code is a bit muddled because I originally wanted the z=0 plane to
 * be exactly in focus. Later I changed my mind and moved the eye to the
 * origin instead, and the focal plane is forward enough so that the screen
 * coordinates work out properly on it
 */
void transscene(void) {
	double tr[3];
	double rot[3][3];
	double scale[3][3];
	double combo[3][3];
	double combo2[3][3];
	double v[3];
	double v2[3];
	int i, j, k, n, a;
	double f, g;

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			rot[i][j] = scale[i][j] = 0;
		}
	}

	// scale x and y coords accordingly, and then scale averagely
	scale[0][0] = s.swidth / s.wwidth;
	scale[1][1] = s.sheight / s.wheight;
	scale[0][0] = scale[1][1] = scale[2][2] = (scale[0][0] + scale[1][1]) / 2;
	printf("Scaling matrix is:\n");
	showmat(scale);
	printf("\n");

	// want the camera to be facing, and focused on the z=0 plane
	// work out what should go to (0, 0, 0)

	subvec(tr, s.at, s.eye);
	subvec(v, s.at, s.eye);
	/*
	showvec(v);
	printf("%lf\n\n", lenvec(v));
	*/
	multvec(tr, 1 - (s.focal / lenvec(v)), tr);
	subvec(tr, tr, s.at);
	printf("Translating by");
	showvec(tr);
	printf("\n");

	// rotate about x-axis so that eye lies in x-z plane
	rot[0][0] = 1;
	f = sqrt(v[2] * v[2] + v[1] * v[1]);
	rot[1][1] = rot[2][2] = -v[2] / f;
	rot[2][1] = -(rot[1][2] = v[1] / f);

	//matmat(combo, scale, rot);
	assignmat(combo, rot);

	/*
	matvec(v2, rot, v);
	printf("Check");showvec(v2);
	*/

	// rotate about y-axis so the eye is on the z-axis
	rot[2][1] = rot[1][2] = 0;
	rot[1][1] = 1;
	g = sqrt(f * f + v[0] * v[0]);
	rot[0][0] = rot[2][2] = -f / g;
	rot[0][2] = -(rot[2][0] = v[0] / g);

	matmat(combo2, rot, combo);

	/*
	matvec(v2, combo2, v);
	printf("Check");showvec(v2);
	*/

	matvec(v, combo2, s.up);

	// rotate about z-axis so s.up is up
	// actually it points down, but the screen coords
	// are upside-down so it cancels out
	rot[0][2] = rot[2][0] = 0;
	rot[2][2] = 1;
	f = sqrt(v[0] * v[0] + v[1] * v[1]);
	rot[0][0] = rot[1][1] = -v[1] / f;
	rot[1][0] = -(rot[0][1] = v[0] / f);

	matmat(combo, rot, combo2);

	/*
	matvec(v2, combo, s.up);
	printf("Check");showvec(v2);
	matvec(v2, combo, s.eye);
	printf("Check");showvec(v2);
	*/

	matmat(combo2, combo, scale);
	// now combo = rotations no scaling
	// combo2 = rotations with scaling

	addvec(v, s.eye, tr);
	matvec(v2, combo2, v);
	printf("Eye now at");
	showvec(v2);
	printf("\n");

	seye = -v2[2];

	/* Not accurate enough for this:
	if (v2[0] != 0.0 || v2[1] != 0.0) {
		fprintf(stderr, "Transformations didn't work!\n");
		showvec(v2);
		exit(1);
	}
	*/

	printf("Adding %lf to z coords\n", seye);

	// Transform all vectors, calculate screen coords after projection
	//printf("New vectors are:\n");
	for (i=0; i<vcount; i++) {
		addvec(v, vertex[i], tr);
		matvec(svertex[i], combo2, v);
		svertex[i][2] += seye;
		//showvec(svertex[i]);
		logprintf("svertex %d:", i);
		logvec(svertex[i]);
		f = svertex[i][0];
		f *= seye / svertex[i][2];
		scoord[i][0] = (int) (f + s.swidth / 2);
		f = svertex[i][1];
		f *= seye / svertex[i][2];
		scoord[i][1] = (int) (f + s.sheight / 2);
	}

	// Transform all lights
	for (i=0; i<lightcount; i++) {
		addvec(v, light[i].loc, tr);
		matvec(light[i].sloc, combo2, v);
		light[i].sloc[2] += seye;
	}

	// Fix up the normals
	// Calculate max, min coords
	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			// Don't translate normals
			// Normals will be wrong if funny scaling is used
			matvec(o[i].face[j].sn, combo, o[i].face[j].n);
			// Stupid sort
			// work out which is the largest component of n
			o[i].face[j].maxn = 0;
			if (fabs(o[i].face[j].sn[1]) > fabs(o[i].face[j].sn[0])) {
				o[i].face[j].maxn = 1;
				if (fabs(o[i].face[j].sn[2]) > fabs(o[i].face[j].sn[1])) {
					o[i].face[j].maxn = 2;
				}
			} else if (fabs(o[i].face[j].sn[2]) > fabs(o[i].face[j].sn[0])) {
				o[i].face[j].maxn = 2;
			}
			// Eventually I'll make this more efficient 
			//multvec(o[i].face[j].sn, 1.0 / lenvec(v), v);
			/*
			subvec(v, svertex[o[i].face[j].e[0][1]], svertex[o[i].face[j].e[0][0]]);
			printf("zero: %lf\n", dotvec(v, o[i].face[j].sn));
			*/
			for (n=0; n<2; n++) {
				o[i].face[j].min[n] = o[i].face[j].max[n] =
					scoord[o[i].face[j].e[0][0]][n];
				for (k=0; k<o[i].face[j].ecount; k++) {
					a = scoord[o[i].face[j].e[k][0]][n];
					if (o[i].face[j].min[n] > a) {
						o[i].face[j].min[n] = a;
					}
					if (o[i].face[j].max[n] < a) {
						o[i].face[j].max[n] = a;
					}
				}
			}

		}
	}

}

void *salloc(size_t size) {
	void *ptr;
	
	ptr = malloc(size);
	if (!ptr) {
		fprintf(stderr, "out of memory\n");
		exit(2);
	}

	return(ptr);
}

// Convert to macros for speed? Or will gcc automatically inline them?
void addvec(double *a, double *b, double *c) {
	a[0] = b[0] + c[0];
	a[1] = b[1] + c[1];
	a[2] = b[2] + c[2];
}

void subvec(double *a, double *b, double *c) {
	a[0] = b[0] - c[0];
	a[1] = b[1] - c[1];
	a[2] = b[2] - c[2];
}

void multvec(double *r, double a, double *v) {
	r[0] = v[0] * a;
	r[1] = v[1] * a;
	r[2] = v[2] * a;
}

double dotvec(double *a, double *b) {
	return(a[0]*b[0] + a[1]*b[1] + a[2]*b[2]);
}

// I hope gcc unrolls these
void matvec(double *r, double a[3][3], double *v) {
	int i, j;

	for (i=0; i<3; i++) {
		r[i] = 0;
		for (j=0; j<3; j++) {
			r[i] += a[i][j] * v[j];
		}
	}
}

void matmat(double a[3][3], double b[3][3], double c[3][3]) {
	int i, j, k;

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			a[i][j] = 0;
			for (k=0; k<3; k++) a[i][j] += b[i][k] * c[k][j];
		}
	}
}

void assignmat(double a[3][3], double b[3][3]) {
	int i, j;

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			a[i][j] = b[i][j];
		}
	}
}

void showvec(double *a) {
	int i;

	for (i=0; i<3; i++) {
		printf(" %lf", a[i]);
	}
	printf("\n");
}

void showmat(double a[3][3]) {
	int i;

	for (i=0; i<3; i++) {
		showvec(a[i]);
	}
}

// See Eric Haine's paper in Glassner's "Introduction to Ray Tracing"
char inface(double v[3], struct facestruct *f) {
	double x, nx;
	double y, ny;
	int a, b;
	int sh, nsh, nc;
	int i;

	// eliminate a coordinate - we find which component of the
	// normal is largest and eliminate this
	// a, b are the coords we keep
	switch(f->maxn) {
		case 0:
			a = 1;
			b = 2;
			break;

		case 1:
			a = 0;
			b = 2;
			break;

		// case 2: (default is faster?)
		default:
			a = 0;
			b = 1;
			break;
	}

	logprintf("normal:");
	logvec(f->sn);
	logprintf("a, b: %d, %d\n", a, b);
	// translate so that v is the origin
	x = svertex[f->e[0][0]][a] - v[a];
	y = svertex[f->e[0][0]][b] - v[b];
	sh = (y >= 0 ? 1 : -1);

	logprintf("vertex: %lf, %lf\n", v[a], v[b]);

	// init no. of crossings
	nc = 0;
	for (i=1; i<f->ecount; i++) {
		logprintf("%d: x, y: %lf, %lf\n", i, x, y);
		nx = svertex[f->e[i][0]][a] - v[a];
		ny = svertex[f->e[i][0]][b] - v[b];
		nsh = (ny >= 0 ? 1 : -1);
		if (nsh != sh) {
			if ((x>0) && (nx>0)) {
				nc++;
			} else if ((x>0) || (nx>0)) {
				if (x - y*(nx-x)/(ny-y) > 0) {
					logprintf("%lf %lf - %lf %lf by case 2\n", x, y, nx, ny);
					logprintf("%lf", x - y*(nx-x)/(ny-y));
					nc++;
				}
			}

		}
		sh = nsh;
		x = nx;
		y = ny;
	}

		logprintf("%d: x, y: %lf %lf\n", i, x, y);
		ny = svertex[f->e[0][0]][b] - v[b];
		nsh = (ny >= 0 ? 1 : -1);
		if (nsh != sh) {
			nx = svertex[f->e[0][0]][a] - v[a];
			if ((x>0) && (nx>0)) {
				logprintf("yes by case 3\n");
				nc++;
			} else if ((x>0) || (nx>0)) {
				if (x - y*(nx-x)/(ny-y) > 0) {
					logprintf("%lf %lf - %lf %lf by case 4\n", x, y, nx, ny);
					nc++;
				}
			}

		}
	
	logprintf("NC: %d\n", nc);
	return (nc % 2);
}

// put faces on active/passive list in order of min y coord
// for now, do a insertion sort
void initfacelist(void) {
	struct facelist *fp, *temp;
	int i, j, n;

	active = (struct facelist *) salloc(sizeof(struct facelist));
	active->next = NULL;
	active->face = NULL;

	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			fp = active;
			n = o[i].face[j].min[1];
			while (fp->next && n > fp->next->face->min[1]) fp = fp->next;
			temp = fp->next;
			fp->next = (struct facelist *) salloc(sizeof(struct facelist));
			fp->next->face = &o[i].face[j];
			fp->next->next = temp;
			
		}
	}

	/*
	fp = active;
	while (fp->next) {
		fp = fp->next;
		printf("%d ", fp->face->min[1]);
	}
	*/
}

void wireframe(void) {
	int i, j;

	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			outline(&o[i].face[j]);
		}
	}
}

void setcolourpt(double *v, struct facestruct *f) {
	int i, j;
	double colour[3];
	double costheta, cosphi, mu;
	double d[3];
	double b[3];
	double s[3];

	for (i=0; i<3; i++) {
		colour[i] = f->colour[i] * amblight[i];
	}

	for (i=0; i<lightcount; i++) {
		subvec(d, light[i].sloc, v);
		mu = 1.0 / lenvec(d);
		multvec(s, mu, d);
		// work out cosine of the angle
		costheta = fabs(dotvec(s, f->sn));
		logprintf("costheta: %lf\n", costheta);
		subvec(b, s, v);
		multvec(b, 1.0 / lenvec(b), b);
		logprintf("b: ");
		logvec(b);
		cosphi = dotvec(b, f->sn);
		cosphi = fabs(2 * cosphi * cosphi - 1);
		logprintf("cosphi: %lf\n", cosphi);
		for (j=0; j<f->specpower; j++) cosphi *= cosphi;
		logprintf("cosphi specpowered: %lf\n", cosphi);
		
		mu = 1000.0 / (lenvec(d) + D0);
		logprintf("lighting mu: %lf\n", mu);

		for (j=0; j<3; j++) {
			colour[j] += mu * f->colour[j] * light[i].colour[j] *
				(f->diffuse * costheta + amblight[j] +
				f->spec * cosphi);
			if (colour[j] > 1) {
				/*
				printf("saturation\n");
				exit(3);
				*/
				colour[j] = 1;
			}
		}
	}
	setcolour(initrgb((int) (colour[0]*255), (int) (colour[1]*255), (int) (colour[2]*255)));
}
